import rope.refactor.importutils

from traad.rope.validate import validate
from traad.state import task_state_monitor
import traad.trace


class ImportUtilsMixin:
    @validate
    @task_state_monitor
    def _importutil_func(self, state, path, funcname):
        # TODO: Update state in some useful way.
        path = self.to_relative_path(path)
        iorg = rope.refactor.importutils.ImportOrganizer(self.proj)
        changes = getattr(iorg, funcname)(self.proj.get_resource(path))
        if changes:
            self.proj.do(changes)


    @traad.trace.trace
    def organize_imports(self, state, path):
        """Organize the import statements in a python source file.

        Args:
        path: The path of the file to reorganize.
        """

        # TODO: This takes more arguments.

        return self._importutil_func(state, path,
                                     "organize_imports")


    @traad.trace.trace
    def expand_star_imports(self, state, path):
        """Expand "star" import statements in a python source file.

        Args:
        path: The path of the file to reorganize.
        """
        return self._importutil_func(state, path,
                                     "expand_star_imports")


    @traad.trace.trace
    def froms_to_imports(self, state, path):
        """Convert "from" imports to normal imports.

        Args:
        path: The path of the file to reorganize.
        """
        return self._importutil_func(state, path,
                                     "froms_to_imports")


    @traad.trace.trace
    def relatives_to_absolutes(self, state, path):
        """Convert relative imports to absolute.

        Args:
        path: The path of the file to reorganize.
        """
        return self._importutil_func(state, path,
                                     "relatives_to_absolutes")


    @traad.trace.trace
    def handle_long_imports(self, state, path):
        """Clean up long import statements.

        Args:
        path: The path of the file to reorganize.
        """
        return self._importutil_func(state, path,
                                     "handle_long_imports")



import itertools
import logging
import sys

import traad.app


log = logging.getLogger('traad.server')

def run_server(app, port):
    host = 'localhost'

    log.info('Python version: {}'.format(sys.version))

    log.info(
        'Running traad server for app "{}" at {}:{}'.format(
            app,
            host,
            port))

    try:
        app.run(host=host, port=port)
    finally:
        stop_app(app)

def main():
    import argparse

    parser = argparse.ArgumentParser(
        description='Run a traad server.')

    parser.add_argument(
        '-p, --port', metavar='N', type=int,
        dest='port', default=0,
        help='the port on which the server will listen. '
             '(0 selects an unused port.)')

    parser.add_argument(
        '-V, --verbosity', metavar='N', type=int,
        dest='verbosity', default=0,
        help='Verbosity level (0=normal, 1=info, 2=debug).')

    parser.add_argument(
        'project', metavar='P', type=str,
        help='the directory containing the project to serve')

    args = parser.parse_args()

    # Configure logging
    level = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG
    }[args.verbosity]

    logging.basicConfig(
        level=level)

    with traad.app.bind_to_project(args.project) as app:
        run_server(app, args.port)

if __name__ == '__main__':
    main()



import time
import unittest

import webtest

import traad.app


class LongRunningTest(unittest.TestCase):
    def test_long_running(self):
        with traad.app.bind_to_project('.') as traad_app:
            app = webtest.TestApp(traad_app)

            resp = app.post_json(
                '/test/long_running',
                {'message': 'Test message!'})

            task_id = resp.json['task_id']

            results = set()

            start_time = time.time()
            while time.time() - start_time < 15:
                resp = app.get('/task/{}'.format(task_id))
                results.add(resp.body)
                time.sleep(0.1)

            self.assertEqual(len(results), 11)

if __name__ == '__main__':
    unittest.main()



import traad.trace
from traad.rope.validate import validate


def _history_info(seq, idx):
    def contents(c):
        return {
            'resource': c.resource.path,
            'change': c.get_description(),
        }

    c = seq[idx]
    return {
        'description': c.description,
        'time': c.time,
        'full_change': c.get_description(),
        'changes': [contents(x) for x in c.changes],
    }


class HistoryMixin:
    @traad.trace.trace
    @validate
    def undo(self, index=0):
        self.proj.history.undo(
            self.proj.history.undo_list[index])


    @traad.trace.trace
    @validate
    def redo(self, idx=0):
        '''Redo the last undone operation.
        '''
        self.proj.history.redo(
            self.proj.history.redo_list[idx])


    @traad.trace.trace
    def undo_history(self):
        '''Get a list of undo-able changes.

        Returns:
          A list of descriptions of undoable changes/refactorings.
        '''
        return [cs.description for cs in self.proj.history.undo_list]


    @traad.trace.trace
    def redo_history(self):
        '''Get a list of redo-able changes.

        Returns:
          A list of descriptions of redoable changes/refactorings.
        '''
        return [cs.description for cs in self.proj.history.redo_list]


    @traad.trace.trace
    def undo_info(self, idx):
        '''Get information about a single undoable operation.

        Args:
          idx: An index in the undo history.

        Raises:
          IndexError: ``idx`` is out of range.

        Returns:
           A dict of information about the undoable change.
        '''
        return _history_info(self.proj.history.undo_list, idx)


    @traad.trace.trace
    def redo_info(self, idx):
        '''Get information about a single redoable operation.

        Args:
          idx: An index in the redo history.

        Raises:
          IndexError: ``idx`` is out of range.

        Returns:
           A dict of information about the redoable change.
        '''
        return _history_info(self.proj.history.redo_list, idx)



import rope.refactor

from traad.rope.validate import validate
from traad.state import task_state_monitor
import traad.trace


class RenameMixin:
    @traad.trace.trace
    @validate
    @task_state_monitor
    def rename(self, state, new_name, path, offset=None):
        path = self.to_relative_path(path)

        # Construct the refactoring object
        ref = self.make_refactoring(
            rope.refactor.rename.Rename,
            self.get_resource(path),
            offset)

        # This gets a change object that knows about the new-name.
        change = ref.get_change(new_name)

        # Update some state. The state is reentrant.
        # Note that `state` encapsulates the task_id for this
        # refactoring, so we don't need to deal with it.
        state.update(
            {'description': list(change.descriptions),
             'changed_resources': [r.name for r in change.resources],
             })

        # actually run the refactoring
        change.perform()



import unittest

import with_fixture

from traad.state import State
from traad.test import common


class CrossTests(with_fixture.TestCase):
    def withFixture(self):
        with common.use_project({'main': ['basic'],
                                 'cross': ['use_bar']}) as self.proj,\
            common.use_proxy(State.start().proxy()) as self.state:

            self.task_id = 1
            self.state.create(self.task_id).get()
            self.task_state = self.state.get_task_state(self.task_id).get()

            yield

    def test_cross_normalize_arguments(self):
        self.proj.normalize_arguments(
            self.task_state,
            'basic/bar.py',
            163).get()

        common.compare_projects(
            'cross_basic_normalize_arguments',
            'main',
            'basic')

        common.compare_projects(
            'cross_use_bar_normalize_arguments',
            'cross',
            'use_bar')

    def test_cross_remove_argument(self):
        self.proj.remove_argument(
            self.task_state,
            1,
            'basic/bar.py',
            163).get()

        common.compare_projects(
            'cross_basic_remove_argument',
            'main',
            'basic')

        common.compare_projects(
            'cross_use_bar_remove_argument',
            'cross',
            'use_bar')

if __name__ == '__main__':
    unittest.main()



import rope.refactor.extract

from traad.rope.validate import validate
from traad.state import task_state_monitor
import traad.trace


class ExtractMixin:
    @traad.trace.trace
    @validate
    @task_state_monitor
    def _extract(self, state, method, name, path, start_offset, end_offset):
        """Shared implementation of extract_method and _variable.
        """
        path = self.to_relative_path(path)

        ref = self.make_refactoring(
            method,
            self.get_resource(path),
            start_offset,
            end_offset)

        change = ref.get_change(name)

        state.update(
            {'description': list(change.descriptions),
             'changed_resources': [r.name for r in change.resources],
         })

        change.perform()


    @traad.trace.trace
    def extract_method(self, state, name, path, start_offset, end_offset):
        """Extract a method.

        ``path`` may be absolute or relative. If ``path`` is relative,
        then it must to be relative to the root of the project.

        Args:
          new_name: The name for the new method.
          path: The path of the resource containing the code.
          start_offset: The starting offset of the region to extract.
          end_offset: The end (one past the last character) of the
            region to extract.
        """
        self._extract(state,
                      rope.refactor.extract.ExtractMethod,
                      name,
                      path,
                      start_offset,
                      end_offset)


    @traad.trace.trace
    def extract_variable(self, state, name, path, start_offset, end_offset):
        """Extract a variable.

        ``path`` may be absolute or relative. If ``path`` is relative,
        then it must to be relative to the root of the project.

        Args:
          new_name: The name for the new variable.
          path: The path of the resource containing the code.
          start_offset: The starting offset of the region to extract.
          end_offset: The end (one past the last character) of the
            region to extract.
        """
        self._extract(state,
                      rope.refactor.extract.ExtractVariable,
                      name,
                      path,
                      start_offset,
                      end_offset)



import itertools
import logging
import sys
import traceback

import decorator

try:
    import reprlib
    repr = reprlib.repr
except ImportError:
    pass

log = logging.getLogger('traad.trace')

@decorator.decorator
def trace(f, *args, **kw):
    '''A simple tracing decorator, mostly to help with debugging.
    '''

    log.info('{}({})'.format(
        f.__name__,
        ', '.join(
            map(repr,
                itertools.chain(
                    args,
                    kw.values())))))

    try:
        return f(*args, **kw)
    except:
        einfo = sys.exc_info()
        log.error('Exception in {}: {}'.format(
            f.__name__,
            ''.join(traceback.format_exception(einfo[0], einfo[1], einfo[2]))))
        raise



import threading
import time

from traad.rope.validate import validate
import traad.trace


lrthreads = set()

class LRThread(threading.Thread):
    def __init__(self, state, message):
        super(LRThread, self).__init__()
        self.state = state
        self.message = message

    def run(self):
        self.state.update({'status': 'starting'})

        time.sleep(1)

        for i in range(10):
            self.state.update({'status{}'.format(i): self.message})
            time.sleep(1)

        lrthreads.discard(self)

@traad.trace.trace
def long_running(state, message):
    lrt = LRThread(state, message)
    lrthreads.add(lrt)
    lrt.start()



import contextlib
import itertools
import logging
import sys

from . import bottle
from .rope.project import Project
from .state import State, TaskState


log = logging.getLogger('traad.app')

PROTOCOL_VERSION = 2


class ProjectApp(bottle.Bottle):
    def __init__(self):
        super(ProjectApp, self).__init__()
        self.project = None
        self.state = None
        self.task_ids = itertools.count()

app = ProjectApp()


@contextlib.contextmanager
def bind_to_project(project_path):
    app.project = Project.start(project_path).proxy()
    app.state = State.start().proxy()

    try:
        yield app
    finally:
        app.state.stop()
        app.project.stop()


@app.get('/protocol_version')
def protocol_version_view():
    return {'protocol-version': PROTOCOL_VERSION}


@app.get('/root')
def project_root_view():
    try:
        return {
            'result': 'success',
            'root': bottle.request.app.project.get_root().get()
        }
    except Exception:
        return 'huh...'


@app.get('/all_resources')
def all_resources():
    return {
        'result': 'success',
        'resources': bottle.request.app.project.get_all_resources().get()
    }


@app.get('/task/<task_id>')
def task_status_view(task_id):
    try:
        return bottle.request.app.state.get_task_state(int(task_id)).get()
    except KeyError:
        bottle.abort(404, "No task with that ID")


@app.get('/tasks')
def full_task_status():
    status = bottle.request.app.state.get_full_state().get()
    log.info('full status: {}'.format(status))
    return status


@app.post('/history/undo')
def undo_view():
    args = bottle.request.json
    bottle.request.app.project.undo(args['index']).get()

    # TODO: What if it actually fails?
    return {'result': 'success'}


@app.post('/history/redo')
def redo_view():
    args = bottle.request.json
    bottle.request.app.project.redo(args['index']).get()

    # TODO: What if it actually fails?
    return {'result': 'success'}


@app.get('/history/view_undo')
def undo_history_view():
    return {
        'result': 'success',
        'history': bottle.request.app.project.undo_history().get()
    }


@app.get('/history/view_redo')
def redo_history_view():
    return {
        'result': 'success',
        'history': bottle.request.app.project.redo_history().get()
    }


@app.get('/history/undo_info/<idx>')
def undo_info_view(idx):
    return {
        'result': 'success',
        'info': bottle.request.app.project.undo_info(int(idx)).get()
    }


@app.get('/history/redo_info/<idx>')
def redo_info_view(idx):
    return {
        'result': 'success',
        'info': bottle.request.app.project.redo_info(int(idx)).get()
    }


@app.post('/test/long_running')
def long_running_test():
    import traad.test.tasks as tasks
    args = bottle.request.json

    return standard_async_task(tasks.long_running,
                               args['message'])


@app.post('/refactor/rename')
def rename_view():
    args = bottle.request.json
    return standard_async_task(bottle.request.app.project.rename,
                               args['name'],
                               args['path'],
                               args.get('offset'))


def extract_core(method, request):
    """Common implementation for extract-method and extract-variable views.

    Args:
      method: The refactoring method to actually call.
      request: The bottle request for the refactoring.
    """
    args = request.json
    return standard_async_task(method,
                               args['name'],
                               args['path'],
                               args['start-offset'],
                               args['end-offset'])


@app.post('/refactor/extract_method')
def extract_method_view():
    return extract_core(bottle.request.app.project.extract_method,
                        bottle.request)


@app.post('/refactor/extract_variable')
def extract_variable_view():
    return extract_core(bottle.request.app.project.extract_variable,
                        bottle.request)


@app.post('/refactor/normalize_arguments')
def normalize_arguments_view():
    args = bottle.request.json
    return standard_async_task(bottle.request.app.project.normalize_arguments,
                               args['path'],
                               args['offset'])


@app.post('/refactor/remove_argument')
def remove_argument_view():
    args = bottle.request.json
    return standard_async_task(bottle.request.app.project.remove_argument,
                               args['arg_index'],
                               args['path'],
                               args['offset'])


@app.post('/code_assist/completions')
def code_assist_completion_view():
    args = bottle.request.json

    log.info('get completion: {}'.format(args))

    with open(args['path'], 'r') as f:
        code = f.read()

    results = bottle.request.app.project.code_assist(code,
                                                     args['offset'],
                                                     args['path']).get()

    # TODO: What if it fails?
    return {
        'result': 'success',
        'completions': results,
    }


@app.post('/code_assist/doc')
def code_assist_doc_view():
    args = bottle.request.json

    log.info('get doc: {}'.format(args))

    with open(args['path'], 'r') as f:
        code = f.read()

    doc = bottle.request.app.project.get_doc(
        code=code,
        offset=args['offset'],
        path=args['path']).get()

    return {
        'result': 'failure' if doc is None else 'success',
        'doc': doc
    }


@app.post('/code_assist/calltip')
def code_assist_calltip_view():
    args = bottle.request.json

    log.info('get calltip: {}'.format(args))

    with open(args['path'], 'r') as f:
        code = f.read()

    calltip = bottle.request.app.project.get_calltip(
        code=code,
        offset=args['offset'],
        path=args['path']).get()

    return {
        'result': 'failure' if calltip is None else 'success',
        'calltip': calltip
    }


# @get('/code_assist/definition')
# def code_assist_definition_view():
#     args = request.json

#     log.info('get definition: {}'.format(args))

#     return {
#         'results': bottle.request.app.project.get_definition_location(
#             code=args['code'],
#             offset=args['offset'],
#             path=args['path'])
#     }


@app.post('/findit/occurrences')
def findit_occurences_view():
    args = bottle.request.json
    data = bottle.request.app.project.find_occurrences(
        args['offset'],
        args['path']).get()

    # TODO: What if it actually fails?
    return {
        'result': 'success',
        'data': data,
    }


@app.post('/findit/implementations')
def findit_implementations_view():
    args = bottle.request.json
    data = bottle.request.app.project.find_implementations(
        args['offset'],
        args['path']).get()

    # TODO: What if it actually fails?
    return {
        'result': 'success',
        'data': data,
    }


@app.post('/findit/definition')
def findit_definitions_view():
    args = bottle.request.json

    with open(args['path'], 'r') as f:
        code = f.read()

    data = bottle.request.app.project.find_definition(
        code,
        args['offset'],
        args['path']).get()

    # TODO: What if it actually fails?
    return {
        'result': 'success',
        'data': data,
    }


def _importutil_core(request, method):
    # TODO: This patterns of async-tasks is repeated several times.
    # Refactor it.

    args = request.json
    return standard_async_task(
        method,
        args['path'])


@app.post("/imports/organize")
def organize_imports_view():
    return _importutil_core(bottle.request,
                            bottle.request.app.project.organize_imports)


@app.post("/imports/expand_star")
def expand_star_imports_view():
    return _importutil_core(bottle.request,
                            bottle.request.app.project.expand_star_imports)


@app.post("/imports/froms_to_imports")
def from_to_imports_view():
    return _importutil_core(bottle.request,
                            bottle.request.app.project.froms_to_imports)


@app.post("/imports/relatives_to_absolutes")
def relatives_to_absolutes_view():
    return _importutil_core(bottle.request,
                            bottle.request.app.project.relatives_to_absolutes)


@app.post("/imports/handle_long_imports")
def handle_long_imports_view():
    return _importutil_core(bottle.request,
                            bottle.request.app.project.handle_long_imports)


def standard_async_task(method, *args):
    """Launch a typical async task.

    This creates a `TaskState` for the new task and runs the task. The
    assumption here is that `method` is a pykka actor method that will
    execute in a separate thread. This function doesn't do anything to
    magically make `method` execute asynchronously.

    Args:
      method: The asynchronous callable to execute.
      args: The arguments to pass to ``method``.

    """
    log.info('{}: {}'.format(method, args))

    try:
        task_id = next(bottle.request.app.task_ids)
        state = bottle.request.app.state
        state.create(task_id)

        method(TaskState(state, task_id),
               *args)

        log.info('{}: success'.format(method))

        return {
            'result': 'success',
            'task_id': task_id
        }
    except:
        e = sys.exc_info()[1]
        log.error('{} error: {}'.format(method, e))
        return {
            'result': 'failure',
            'message': str(e)
        }



import timeit

from traad.test.acceptance.json_api import json_request

filename = '/Users/abingham/projects/pylons/pyramid/pyramid/request.py'

with open(filename, 'rt') as f:
    data = f.read()

def func():
    json_request('http://localhost:23445/code_assist/doc',
                 {'offset': 4683,
                  'path': filename,
                  'code': data},
                 method='GET')

print(
    timeit.timeit('func()', setup='from __main__ import func', number=100))


import sys

import rope.refactor.change_signature

from traad.rope.validate import validate
from traad.state import task_state_monitor
import traad.trace


class ChangeSignatureMixin:
    @validate
    @task_state_monitor
    def change_sig(self, state, path, offset, refactoring):
        """Common implementation for change-signature refactorings.

        ``path`` may be absolute or relative. If ``path`` is relative,
        then it must to be relative to the root of the project.

        Args:
          self: The Project on which this operates.
          state: The TaskState for this refactoring.
          path: The path of the file/directory to query.
          offset: The offset in the resource of the method signature.
          refactoring: The refactoring job to run.
        """

        path = self.to_relative_path(path)

        ref = self.make_refactoring(
            rope.refactor.change_signature.ChangeSignature,
            self.get_resource(path),
            offset)

        change = ref.get_change(
            [refactoring])

        state.update(
            {'description': list(change.descriptions),
             'changed_resources': [r.name for r in change.resources],
             })

        change.perform()

    @traad.trace.trace
    def normalize_arguments(self, state, path, offset):
        """Normalize arguments for a method.

        ``path`` may be absolute or relative. If ``path`` is relative,
        then it must to be relative to the root of the project.

        Args:
          self: The Project on which this operates.
          state: The TaskState for this refactoring.
          path: The path of the file/directory to query.
          offset: The offset in the resource of the method signature.
        """
        self.change_sig(
            state,
            path,
            offset,
            rope.refactor.change_signature.ArgumentNormalizer())


    @traad.trace.trace
    def remove_argument(self,
                        state,
                        arg_index,
                        path,
                        offset):
        """Remove an argument from a method.

        ``path`` may be absolute or relative. If ``path`` is relative,
        then it must to be relative to the root of the project.

        Args:
          self: The Project on which this operates.
          state: The TaskState for this refactoring.
          arg_index: The index of the argument to remove.
          path: The path of the file/directory to query.
          offset: The offset in the resource of the method signature.
        """
        self.change_sig(
            state,
            path,
            offset,
            rope.refactor.change_signature.ArgumentRemover(arg_index))



import json
import os
import time
import unittest

import webtest
import with_fixture

import traad.app
from traad.test import common


def wait_for_task(task_id, app):
    while True:
        resp = app.get('/task/{}'.format(task_id))

        if resp.json['status'] == 'success':
            return True
        elif resp.json['status'] == 'failure':
            return False

        time.sleep(0.01)


class JSONAPITests(with_fixture.TestCase):
    def withFixture(self):
        common.activate({'main': ['basic']})

        with traad.app.bind_to_project(common.activated_path('main')) as self.traad_app:
            self.app = webtest.TestApp(self.traad_app)

            yield

        common.deactivate()

    def test_rename(self):
        resp = self.app.post_json(
            '/refactor/rename',
            {
                'name': 'Llama',
                'path': 'basic/foo.py',
                'offset': 8,
            })

        self.assertEqual(resp.json['result'], 'success')
        task_id = resp.json['task_id']

        self.assertTrue(
            wait_for_task(task_id, self.app))

        common.compare_projects(
            'basic_rename_llama',
            'main',
            'basic')

    def test_find_occurrences(self):
        resp = self.app.post_json(
            '/findit/occurrences',
            {
                'path': 'basic/foo.py',
                'offset': 8,
            })
        self.assertEqual(len(resp.json['data']), 3)

    def test_find_implementations(self):
        resp = self.app.post_json(
            '/findit/implementations',
            {
                'path': 'basic/overrides.py',
                'offset': 33,
            })
        self.assertEqual(len(resp.json['data']), 1)

    def test_find_definition(self):
        path = os.path.join(
            common.activated_path('main'),
            'basic', 'bar.py')

        resp = self.app.post_json(
            '/findit/definition',
            {
                'path': path,
                'offset': 142,
            })

        self.assertEqual(
            resp.json['data'],
            [os.path.join('basic', 'bar.py'),
             [91, 100],
             91,
             False,
             7])

    def test_undo_undoes_changes(self):
        resp = self.app.post_json(
            '/refactor/rename',
            {
                'name': 'Llama',
                'path': 'basic/foo.py',
                'offset': 8,
            })

        if resp.json['result'] != 'success':
            print(resp.json['message'])
        self.assertEqual(resp.json['result'], 'success')
        task_id = resp.json['task_id']

        self.assertTrue(
            wait_for_task(task_id, self.app))

        with self.assertRaises(ValueError):
            common.compare_projects(
                'basic',
                'main',
                'basic')

        resp = self.app.post_json(
            '/history/undo',
            {'index': 0})

        self.assertEqual(resp.json['result'], 'success')

        common.compare_projects(
            'basic',
            'main',
            'basic')

if __name__ == '__main__':
    unittest.main()



import logging

import decorator

from traad.rope.log import log

@decorator.decorator
def validate(f, self, *args, **kwargs):
    "Call self.proj.validate() before running the function."
    log.info('Validating')
    self.proj.validate()
    log.info('Done validating')
    return f(self, *args, **kwargs)



import contextlib
import os

from eagertools import emap

import pykka

import rope.base.project
from rope.refactor import multiproject

from traad.rope.change_signature import ChangeSignatureMixin
from traad.rope.codeassist import CodeAssistMixin
from traad.rope.extract import ExtractMixin
from traad.rope.findit import FinditMixin
from traad.rope.history import HistoryMixin
from traad.rope.importutil import ImportUtilsMixin
from traad.rope.rename import RenameMixin


def get_all_resources(proj):
    '''Generate a sequence of (path, is_folder) tuples for all
    resources in a project.

    Args:
      proj: The rope Project to scan.

    Returns: An iterable over all resources in a Project, with a tuple
      (path, is_folder) for each.
    '''
    todo = ['']
    while todo:
        res_path = todo[0]
        todo = todo[1:]
        res = proj.get_resource(res_path)
        yield(res.path, res.is_folder())

        if res.is_folder():
            todo.extend((child.path for child in res.get_children()))


class Change:
    """This represents a single, fully-specified change that can be
    performed.

    This includes both the refactoring type as well as the arguments
    to the refactoring, the resources, etc. You can direcly call
    perform() on this object to run the refactoring.
    """
    def __init__(self, refactoring, *args):
        self.refactoring = refactoring
        self.args = args

        self.changes = self.refactoring.get_all_changes(*args)
        self._performed = False

    @property
    def descriptions(self):
        """An iterable of descriptions of the changes that this
        refactoring will make.
        """
        for proj, cset in self.changes:
            yield cset.get_description()

    @property
    def resources(self):
        """An iterable of resources that this refactoring will modify.
        """
        for proj, cset in self.changes:
            for res in cset.get_changed_resources():
                yield res

    def perform(self):
        "Perform the refactoring."
        assert not self._performed

        multiproject.perform(self.changes)
        self._performed = True


class MultiProjectRefactoring:
    """Support class for performing multi-project refactorings.
    """
    def __init__(self, project, refactoring_type, *args):
        cross_ref = multiproject.MultiProjectRefactoring(
            refactoring_type,
            list(project.cross_projects.values()))
        self.rope_ref = cross_ref(project.proj, *args)

    def get_change(self, *args):
        return Change(self.rope_ref, *args)


class Project(ChangeSignatureMixin,
              CodeAssistMixin,
              ExtractMixin,
              FinditMixin,
              HistoryMixin,
              ImportUtilsMixin,
              RenameMixin,
              pykka.ThreadingActor):
    """An actor that controls access to an underlying Rope project.
    """
    def __init__(self, project_dir, cross_project_dirs=[]):
        super(Project, self).__init__()

        self.proj = rope.base.project.Project(project_dir)

        self.cross_projects = dict()

        cross_dirs = set(cross_project_dirs)
        cross_dirs.discard(project_dir)
        emap(self.add_cross_project, cross_dirs)

    def close(self):
        self.proj.close()

    def add_cross_project(self, directory):
        """Add a cross project rooted at `directory`."""
        self.cross_projects[directory] = rope.base.project.Project(directory)

    def remove_cross_project(self, directory):
        """Remove the cross project rooted at `directory`."""
        del self.cross_projects[directory]

    def cross_project_directories(self):
        """Get a list of root directories for all cross projects."""
        return list(self.cross_projects.keys())

    def to_relative_path(self, path):
        '''Get a version of a path relative to the project root.

        If ``path`` is already relative, then it is unchanged. If
        ``path`` is absolute, then it is made relative to the project
        root.

        Args:
          path: The path to make relative.

        Returns: ``path`` relative to the project root.

        '''
        if os.path.isabs(path):
            path = os.path.relpath(
                os.path.realpath(path),
                self.proj.root.real_path)
        return path

    def get_resource(self, path):
        return self.proj.get_resource(path)

    def make_refactoring(self, refactoring_type, *args):
        return MultiProjectRefactoring(
            self,
            refactoring_type,
            *args)

    def get_children(self, path):
        '''Get a list of all child resources of a given path.

        ``path`` may be absolute or relative. If ``path`` is relative,
        then it must to be relative to the root of the project.

        Args:
          path: The path of the file/directory to query.

        Returns: A list of tuples of the form (path,
          is_folder).

        '''

        path = self._to_relative_path(path)

        children = self.proj.get_resource(path).get_children()
        return [(child.path, child.is_folder()) for child in children]

    def get_all_resources(self):
        '''Get a list of all resources in the project.

        Returns: A list of tuples of the form (path,
            is_folder).
        '''
        return list(get_all_resources(self.proj))

    def get_root(self):
        return self.proj.root.real_path

    def __repr__(self):
        return 'Project("{}")'.format(
            self.proj.root.real_path)

    def __str__(self):
        return repr(self)



import unittest

import with_fixture

from traad.state import State, TaskState
from traad.test import common


class Tests(with_fixture.TestCase):
    def withFixture(self):
        with common.use_project({'main': ['basic']}) as self.proj,\
             common.use_proxy(State.start().proxy()) as self.state:
            self.state.create(1).get()
            self.task_state = TaskState(self.state, 1)
            yield

    def test_simple(self):
        self.proj.rename(
            self.task_state,
            'Llama',
            'basic/foo.py',
            8).get()

        common.compare_projects(
            'basic_rename_llama',
            'main',
            'basic')

if __name__ == '__main__':
    unittest.main()



import unittest


class ProjectTests(unittest.TestCase):
    @unittest.skip('TODO')
    def test_get_all_resources(self):
        self.assertEqual(
            sorted(self.ri.get_all_resources()),
            [('', True),
             ('basic', True),
             ('basic/__init__.py', False),
             ('basic/bar.py', False),
             ('basic/foo.py', False),
             ('basic/overrides.py', False)])

    @unittest.skip('TODO')
    def test_get_children(self):
        self.assertEqual(
            sorted(self.ri.get_children('basic')),
            [('basic/__init__.py', False),
             ('basic/bar.py', False),
             ('basic/foo.py', False),
             ('basic/overrides.py', False)])



from .foo import Foo

class Bar:
    def __init__(self):
        self.foo = Foo()

    def some_func(self):
        return 1

b = Bar()
b.some_func()

def a_free_func(a,    b,

            c):
    return a + b + c



import rope.contrib.codeassist

import traad.trace
from traad.rope.validate import validate


class CodeAssistMixin:
    @traad.trace.trace
    @validate
    def code_assist(self, code, offset, path):
        '''Get code-assist completions for a point in a file.

        ``path`` may be absolute or relative. If ``path`` is relative,
        then it must to be relative to the root of the project.

        Args:
          self: The Project to use.
          code: The source code in which the completion should
            happen. Note that this may differ from the contents of the
            resource at ``path``.
          offset: The offset into ``code`` where the completion should
            happen.
          path: The path to the resource in which the completion is
            being done.

        Returns: A list of tuples of the form (name, documentation,
          scope, type) for each possible completion.
        '''

        path = self.to_relative_path(path)
        results = rope.contrib.codeassist.code_assist(
            self.proj,
            code,
            offset,
            self.get_resource(path))
        rslt = [(r.name, r.get_doc(), r.scope, r.type) for r in results]
        return rslt


    @traad.trace.trace
    @validate
    def get_calltip(self, code, offset, path):
        """Get the calltip of a function.

        ``path`` may be absolute or relative. If ``path`` is relative,
        then it must be relative to the root of the project.

        Args:
          self: The Project to use.
          code: The source code.
          offset: An offset into ``code`` of the object to query.
          path: The path to the resource in which the search is
            being done.

        Returns: A calltip string.

        """

        path = self.to_relative_path(path)
        return rope.contrib.codeassist.get_calltip(
            self.proj,
            code,
            offset,
            self.get_resource(path))


    @traad.trace.trace
    @validate
    def get_doc(self, code, offset, path):
        '''Get docstring for an object.

        ``path`` may be absolute or relative. If ``path`` is relative,
        then it must to be relative to the root of the project.

        Args:
          self: The Project to work against.
          code: The source code.
          offset: An offset into ``code`` of the object to query.
          path: The path to the resource in which the completion is
            being done.

        Returns: The docstring for the object, or None if there is no such
            documentation.
        '''

        path = self.to_relative_path(path)
        return rope.contrib.codeassist.get_doc(
            self.proj,
            code,
            offset,
            self.get_resource(path))


# class CodeAssistFunctions:
#     """The codeassist related functions of the rope interface.

#     A base for RopeInterface.

#     """

#     @traad.trace.trace
#     @validate
#     def get_definition_location(self, code, offset, path):
#         '''Get location of definition for a symbol.

#         ``path`` may be absolute or relative. If ``path`` is relative,
#         then it must to be relative to the root of the project.

#         Args:
#           code: The source code.
#           offset: An offset into ``code`` of the object to query.
#           path: The path to the resource in which the search is
#             being done.

#         Returns: A tuple of the form (path, lineno). If no definition
#           can be found, then (None, None) is returned.
#         '''
#         path = self._to_relative_path(path)
#         rslt = rope.contrib.codeassist.get_definition_location(
#             self.proj,
#             code,
#             offset,
#             self.proj.get_resource(path))

#         if rslt[1] is None:
#             return rslt

#         if rslt[0] is None:
#             return (path, rslt[1])

#         return (rslt[0].real_path, rslt[1])



from eagertools import emap
import rope.contrib.findit

import traad.trace
from traad.rope.validate import validate


def location_to_tuple(location):
    return (location.resource.path,
            location.region,
            location.offset,
            location.unsure,
            location.lineno)


class FinditMixin:
    def _find_locations(self, func, path, offset):
        """Common implementation for occurrences and
        implementations.

        """
        path = self.to_relative_path(path)
        results = func(
            self.proj,
            self.proj.get_resource(path),
            offset)
        return emap(location_to_tuple, results)


    @traad.trace.trace
    @validate
    def find_occurrences(self, offset, path):
        """Find occurrences of a symbol at a point in a file.

        ``path`` may be absolute or relative. If ``path`` is relative,
        then it must to be relative to the root of the project.

        Args:
          self: The traad Project object.
          offset: The offset into ``path`` of the symbol.
          path: The path to the resource containing the symbol to
            search for.

        Returns: A sequence of tuples of the form (path, (region-start,
          region-stop), offset, unsure, lineno).
        """

        return self._find_locations(
            rope.contrib.findit.find_occurrences,
            path,
            offset)


    @traad.trace.trace
    @validate
    def find_implementations(self, offset, path):
        """Find the places a given method is overridden.

        ``path`` may be absolute or relative. If ``path`` is relative,
        then it must to be relative to the root of the project.

        Args:
          self: The traad Project object.
          offset: The offset into ``path`` of the method name.
          path: The path to the resource containing the method name to
            search for.

        Returns: A sequence of tuples of the form (path, (region-start,
          region-stop), offset, unsure, lineno).
        """

        return self._find_locations(
            rope.contrib.findit.find_implementations,
            path,
            offset)


    @traad.trace.trace
    @validate
    def find_definition(self, code, offset, path):
        """Find the definition location of a symbol.

        ``path`` may be absolute or relative. If ``path`` is relative,
        then it must to be relative to the root of the project.

        Args:
          self: The traad Project object.
          code: The source code containing the method symbol.
          offset: The offset into ``code`` of the symbol.
          path: The path to the resource containing ``code``.

        Returns: A tuple of the form (path, (region-start,
          region-stop), offset, unsure, lineno), or `None` if the
          definition can't be found.

        """

        path = self.to_relative_path(path)
        defn = rope.contrib.findit.find_definition(
            self.proj,
            code,
            offset,
            self.get_resource(path))

        return None if defn is None else location_to_tuple(defn)



import contextlib
from filecmp import dircmp
from itertools import chain
import os
import shutil
import sys

try:
    from io import StringIO
except ImportError:
    from StringIO import StringIO

from traad.rope.project import Project


THIS_DIR = os.path.abspath(os.path.split(__file__)[0])
ACTIVE_DIR = os.path.join(THIS_DIR, 'active')
PROJECT_DIR = os.path.join(THIS_DIR, 'projects')

def activate(projects,
             active_dir=ACTIVE_DIR,
             project_dir=PROJECT_DIR):
    """Copy projects from the ``project_dir`` to the ``active_dir``.

    ``projects`` is a dict mapping top-level names in the active
    directory to projects that go there. So, for example::

      {'main' : ['foo', bar']}

    would copy ``project_dir/foo`` and ``project_dir/bar`` to
    ``active_dir/main``.
    """
    try:
        shutil.rmtree(active_dir)
    except OSError:
        pass

    os.makedirs(active_dir)

    for tl_dir, prjs in projects.items():
        tl_dir = os.path.join(active_dir, tl_dir)

        try:
            os.makedirs(tl_dir)
        except OSError:
            pass

        for prj in prjs:
            shutil.copytree(
                os.path.join(project_dir, prj),
                os.path.join(tl_dir, prj))


def activate_project(projects,
                     main_dir='main',
                     active_dir=ACTIVE_DIR,
                     project_dir=PROJECT_DIR):
    activate(projects, active_dir, project_dir)

    return Project.start(
        os.path.join(active_dir, main_dir),
        cross_project_dirs=[os.path.join(active_dir, n)
                            for n in projects if n != main_dir]).proxy()


def activated_path(top_level_name):
    return os.path.join(ACTIVE_DIR, top_level_name)


def deactivate(active_dir=ACTIVE_DIR):
    shutil.rmtree(active_dir)

@contextlib.contextmanager
def use_project(projects,
                main_dir='main',
                active_dir=ACTIVE_DIR,
                project_dir=PROJECT_DIR):
    """A context-manager that calls `activate_project`, yields the
    project, and then on exit stops and deactivates the project.
    """
    try:
        proj = activate_project(projects, main_dir, active_dir, project_dir)
        yield proj
    finally:
        proj.stop()
        deactivate(active_dir)

@contextlib.contextmanager
def use_proxy(proxy):
    """A context-manager to simplify using Pykka actors in with-statement.

    This yields `proxy` and then stops it on exit.
    """
    try:
        yield proxy
    finally:
        proxy.stop()


def diff_report(dc):
    '''Generate a report of differences.

    Stolen directly from `filecmp.dircmp` and modified to be non-insane.

    Args:
        dc: A `dircmp` object.

    Returns: a string with the report.
    '''
    # Output format is purposely lousy
    sio = StringIO()

    if dc.left_only:
        dc.left_only.sort()
        sio.write(u' '.join(chain(['Only in', dc.left, ':'], dc.left_only)))
    if dc.right_only:
        dc.right_only.sort()
        sio.write(u' '.join(chain(['Only in', dc.right, ':'], dc.right_only)))
    if dc.same_files:
        dc.same_files.sort()
        sio.write(u' '.join(chain(['Identical files:'], dc.same_files)))
    if dc.diff_files:
        dc.diff_files.sort()
        sio.write(u' '.join(chain(['Differing files:'], dc.diff_files)))
    if dc.funny_files:
        dc.funny_files.sort()
        sio.write(u' '.join(chain(['Trouble with common files:'], dc.funny_files)))
    if dc.common_dirs:
        dc.common_dirs.sort()
        sio.write(u' '.join(chain(['Common subdirectories:'], dc.common_dirs)))
    if dc.common_funny:
        dc.common_funny.sort()
        sio.write(u' '.join(chain(['Common funny cases:'], dc.common_funny)))

    sio.seek(0)
    return sio.read()

def compare_projects(canned_project,
                     top_level_dir,
                     project,
                     active_dir=ACTIVE_DIR,
                     project_dir=PROJECT_DIR):
    dc = dircmp(os.path.join(project_dir,
                             canned_project),
                os.path.join(active_dir,
                             top_level_dir,
                             project))

    if any([dc.left_only,
            dc.right_only,
            dc.diff_files]):
        dr = diff_report(dc)
        raise ValueError(dr)



import logging
import os
import unittest


def run():
    suite = unittest.TestLoader().discover(
        start_dir=os.path.split(__file__)[0],
        pattern='*_test.py')

    unittest.TextTestRunner(verbosity=1).run(suite)

if __name__ == '__main__':
    logging.basicConfig(
        level=logging.CRITICAL)
    run()



import unittest

import with_fixture

from traad.rope import history
from traad.state import State, TaskState
from traad.test import common


class HistoryTests(with_fixture.TestCase):
    def withFixture(self):
        with common.use_project({'main': ['basic']}) as self.proj,\
             common.use_proxy(State.start().proxy()) as self.state:
            self.state.create(1).get()
            self.task_state = TaskState(self.state, 1)
            yield

    def test_undo_undoes_changes(self):
        self.proj.rename(
            self.task_state,
            'Llama',
            'basic/foo.py',
            8).get()

        with self.assertRaises(ValueError):
            common.compare_projects(
                'basic',
                'main',
                'basic')

        self.proj.undo().get()

        common.compare_projects(
            'basic',
            'main',
            'basic')

    def test_undo_exceptions(self):
        with self.assertRaises(IndexError):
            self.proj.undo().get()

        self.proj.rename(
            self.task_state,
            'Llama',
            'basic/foo.py',
            8).get()

        self.proj.undo().get()

        with self.assertRaises(IndexError):
            self.proj.undo(1).get()

    def test_undo_adds_history(self):
        self.assertEqual(len(self.proj.proj.get().history.undo_list), 0)
        self.proj.rename(
            self.task_state,
            'Llama',
            'basic/foo.py',
            8).get()
        self.assertEqual(len(self.proj.proj.get().history.undo_list), 1)

    def test_redo_redoes_changes(self):
        self.proj.rename(
            self.task_state,
            'Llama',
            'basic/foo.py',
            8).get()

        with self.assertRaises(ValueError):
            common.compare_projects(
                'basic',
                'main',
                'basic')

        self.proj.undo().get()

        common.compare_projects(
            'basic',
            'main',
            'basic')

        self.proj.redo().get()

        with self.assertRaises(ValueError):
            common.compare_projects(
                'basic',
                'main',
                'basic')

    def test_redo_adds_history(self):
        self.proj.rename(
            self.task_state,
            'Llama',
            'basic/foo.py',
            8).get()
        self.assertEqual(len(self.proj.proj.get().history.redo_list), 0)
        self.assertEqual(len(self.proj.proj.get().history.undo_list), 1)
        self.proj.undo().get()
        self.assertEqual(len(self.proj.proj.get().history.redo_list), 1)
        self.assertEqual(len(self.proj.proj.get().history.undo_list), 0)
        self.proj.redo().get()
        self.assertEqual(len(self.proj.proj.get().history.redo_list), 0)
        self.assertEqual(len(self.proj.proj.get().history.undo_list), 1)

    def test_redo_exceptions(self):
        with self.assertRaises(IndexError):
            self.proj.redo().get()

        self.proj.rename(
            self.task_state,
            'Llama',
            'basic/foo.py',
            8).get()

        self.proj.undo().get()
        self.proj.redo().get()

        with self.assertRaises(IndexError):
            self.proj.redo(1).get()

    def test_undo_history(self):
        self.assertEqual(
            len(self.proj.undo_history().get()), 0)
        self.proj.rename(self.task_state,
               'Llama',
               'basic/foo.py',
               8).get()
        self.assertEqual(
            len(self.proj.undo_history().get()), 1)

    def test_undo_info(self):
        self.proj.rename(self.task_state,
               'Llama',
               'basic/foo.py',
               8).get()
        i = self.proj.undo_info(0).get()
        for k in ['description', 'time', 'full_change', 'changes']:
            self.assertIn(k, i)

    def test_undo_info_exceptions(self):
        with self.assertRaises(IndexError):
            self.proj.undo_info(0).get()

        self.proj.rename(self.task_state,
               'Llama',
               'basic/foo.py',
               8).get()
        self.proj.undo_info(0).get()
        with self.assertRaises(IndexError):
            self.proj.undo_info(1).get()

    def test_redo_history(self):
        self.assertEqual(
            len(self.proj.redo_history().get()), 0)
        self.proj.rename(self.task_state,
               'Llama',
               'basic/foo.py',
               8).get()
        self.proj.undo().get()
        self.assertEqual(
            len(self.proj.redo_history().get()), 1)

    def test_redo_info(self):
        self.proj.rename(self.task_state,
               'Llama',
               'basic/foo.py',
               8).get()
        self.proj.undo().get()
        i = self.proj.redo_info(0).get()
        for k in ['description', 'time', 'full_change', 'changes']:
            self.assertIn(k, i)

    def test_redo_info_exceptions(self):
        with self.assertRaises(IndexError):
            self.proj.redo_info(0).get()

        self.proj.rename(self.task_state,
               'Llama',
               'basic/foo.py',
               8).get()
        self.proj.undo().get()

        self.proj.redo_info(0)

        with self.assertRaises(IndexError):
            self.proj.redo_info(1).get()

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.CRITICAL)
    unittest.main()



import sys

from setuptools import setup, find_packages

script_name, rope_lib = {
    2: ('traad', 'rope'),
    3: ('traad3', 'rope_py3k'),
}[sys.version_info.major]

setup(
    name = 'traad',
    version = '0.11',
    packages = find_packages(),

    # metadata for upload to PyPI
    author = 'Austin Bingham',
    author_email = 'austin.bingham@gmail.com',
    description = 'A JSON+HTTP server for the rope Python refactoring library.',
    license = 'MIT',
    keywords = 'refactoring',
    url = 'http://github.com/abingham/traad',

    entry_points = {
        'console_scripts': [
            '{} = traad.server:main'.format(script_name),
            ],
        },

    install_requires=[
        'decorator',
        'eagertools',
        'nose',
        'pykka',
        rope_lib,
        'webtest',
        'with_fixture',
    ],
)



import sys

import decorator
import pykka


class State(pykka.ThreadingActor):
    """Maps task-ids to dicts of data about the task.

    This is essentially a dictionary of task-ids to dicts. Each dict
    describes the state of the task. `State` doesn't care about the
    contents of the dict.

    `State` is a pykka actor and thereby provides thread-save access
    to the task data.
    """
    def __init__(self):
        super(State, self).__init__()

        self.data = {}

    def create(self, task_id, initial_data=None):
        """Create a new entry in the state map for `task-id`.
        """
        if initial_data is None:
            initial_data = {'status': 'pending'}

        assert task_id not in self.data
        self.data[task_id] = initial_data

    def get_task_state(self, task_id):
        """Get the task state for `task-id`.
        """
        return self.data[task_id]

    def set_task_state(self, task_id, data):
        """Replace the task state for `task-id`.
        """
        self.data.task_id = dict(data)

    def update_task_state(self, task_id, data):
        """Update the task state for `task-id`.

        See dict.update() for details on what this means.
        """
        d = dict(self.data[task_id])
        d.update(data)
        self.data[task_id] = d

    def get_full_state(self):
        """Returns a complete copy of the current state.
        """
        # TODO: Should this use deepcopy? probably.
        return dict(self.data)

class TaskState:
    """A wrapper around `State` that provides access to just the state for
    particular task-id.

    """
    def __init__(self, state, task_id):
        self.state = state
        self.task_id = task_id

    def get(self):
        return self.state.get_task_satte(self.task_id).get()

    def set(self, data):
        self.state.set_task_state(self.task_id, dict(data)).get()

    def update(self, data):
        return self.state.update_task_state(self.task_id, data).get()

@decorator.decorator
def task_state_monitor(f, self, task_state, *args, **kwargs):
    """A decorator that updates a `TaskState` to reflect success or
    failure of an operation.

    This executes `f`, passing it `self`, `task_state`, and all of the
    other arguments. If `f` runs without an exception, this updates
    `task_state` with {'status': 'success'} and returns `f`\'s return
    value. If `f` throws, this updates `task_state` with {'status':
    'failure', 'message': <exception info>} and reraises the
    exception.

    This is primarily intended for use on asynchronous calls that need
    to update the state based on their success or failure.

    """
    try:
        r = f(self, task_state, *args, **kwargs)
        task_state.update({'status': 'success'})
        return r
    except:
        task_state.update(
            {'status': 'failure',
             'message': str(sys.exc_info()[1]),
         })
        raise



import os
import unittest

import with_fixture

from traad.test import common


class FinditTests(with_fixture.TestCase):
    def withFixture(self):
        with common.use_project({'main': ['basic']}) as self.proj:
            yield

    def test_find_occurrences(self):
        # Find occurrences of the Foo class
        occ = self.proj.find_occurrences(
            8,
            'basic/foo.py').get()
        self.assertEqual(len(occ), 3)

    def test_find_implementations(self):
        impls = self.proj.find_implementations(
            33,
            'basic/overrides.py').get()
        self.assertEqual(len(impls), 1)

    def test_find_definition(self):
        path = os.path.join(
            common.activated_path('main'),
            'basic', 'bar.py')
        with open(path, 'r') as f:
            code = f.read()
        loc = self.proj.find_definition(
            code,
            142,
            os.path.join('basic', 'bar.py')).get()
        self.assertEqual(
            loc,
            (os.path.join('basic', 'bar.py'),
             (91, 100), 91, False, 7))

if __name__ == '__main__':
    unittest.main()
